import urllib.request
from pathlib import Path


def _download_data(url, overwrite=False):
    data_dir = Path.home() / ".budgetnlp" / "data"
    data_dir.mkdir(parents=True, exist_ok=True)

    filename = Path(url).name
    file_path = data_dir / filename

    if file_path.exists() and not overwrite:
        return file_path

    print(f"Downloading {filename} dictionary to {file_path}")
    urllib.request.urlretrieve(url, file_path)
    print(file_path)
    return file_path


import math

def calculate_drift_score(embeddings, token_counts=None):
    if len(embeddings) < 2:
        raise ValueError("Need at least 2 embeddings to calculate drift")
    
    if token_counts and len(token_counts) != len(embeddings):
        raise ValueError("token_counts must match embeddings length")
    
    results = []
    
    for i in range(len(embeddings) - 1):
        prior_emb = embeddings[i]
        current_emb = embeddings[i + 1]
        
        # Cosine similarity
        dot_product = sum(a * b for a, b in zip(prior_emb, current_emb))
        norm_prior = math.sqrt(sum(a * a for a in prior_emb))
        norm_current = math.sqrt(sum(b * b for b in current_emb))
        
        cosine_sim = dot_product / (norm_prior * norm_current)
        cosine_dist = 1 - cosine_sim
        
        # Drift score (with length penalty if provided)
        drift_score = cosine_dist
        if token_counts:
            drift_score = cosine_dist * math.log(token_counts[i + 1])
        
        results.append({
            'comparison': f'{i} vs {i+1}',
            'cosine_distance': cosine_dist,
            'drift_score': drift_score
        })
    
    return results


# Usage:
# embeddings = [emb_2020, emb_2021, emb_2022, emb_2023]
# token_counts = [5000, 5200, 4800, 6000]
# drift = calculate_drift_score(embeddings, token_counts)


import unicodedata
from functools import wraps

def _string_to_ascii(string):
    return unicodedata.normalize('NFKD', string).encode('ascii', errors='ignore').decode('ascii')

def handle_data_tuples(func):
    """
    Decorator to handle string or list of tuples.
    
    Inputs:
    - string: processes directly, returns [result]
    - [(id, type, content, level), ...]: list of data tuples
    
    Returns:
    - string -> [result]
    - list -> [result1, None, result2, ...] where None = non-'text' types
    """
    @wraps(func)
    def wrapper(data, *args, **kwargs):
        # Check if it's a list of tuples
        if isinstance(data, list):
            results = []
            for item in data:
                if not (isinstance(item, tuple) and len(item) == 4):
                    raise ValueError(f"Expected tuple of length 4, got {type(item)}")
                
                id_, type_, content, level = item
                
                if type_ == 'text':
                    result = func(content, *args, **kwargs)
                    results.append(result)
                else:
                    results.append(None)
            
            return results
        
        # Otherwise treat as string
        else:
            return func(data, *args, **kwargs)
    
    return wrapper


import csv

from flashtext import KeywordProcessor
from .utils.download_data import _download_data
from .utils.text_normalization import _string_to_ascii, handle_data_tuples


url = "https://raw.githubusercontent.com/john-friedman/budgetnlp/main/data/loughran_mcdonald.csv"
file_path = _download_data(url, overwrite=False)

NEGATIONS = {'not', 'no', 'never', 'none', 'neither', 'nobody', 'nothing',
             'nowhere', 'cannot', "can't", "won't", "don't", "doesn't", 
             "didn't", "isn't", "aren't", "wasn't", "weren't", "without"}

NEGATIVE_WORDS = set()

with open(file_path, 'r', encoding='utf-8') as f:
    reader = csv.DictReader(f)
    for row in reader:
        if int(row['Negative']) > 0:
            NEGATIVE_WORDS.add(row['Word'].lower())

processor = KeywordProcessor(case_sensitive=False)
for word in NEGATIVE_WORDS:
    processor.add_keyword(word)





@handle_data_tuples
def negative_ratio(string, negation_reversals=True):
    string = _string_to_ascii(string)
    string_lower = string.lower()
    
    negative_matches = processor.extract_keywords(string_lower)
    negative_count = len(negative_matches)
    
    tokens = string_lower.split()
    total_words = len(tokens)
    
    if negation_reversals:
        reversals = 0
        for i in range(len(tokens) - 1):
            if tokens[i] in NEGATIONS and tokens[i + 1] in NEGATIVE_WORDS:
                reversals += 1
        
        negative_count -= reversals
    
    return negative_count / max(total_words,1)

